/*
 * Copyright (C) 2002-2005 Stefan Stiller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.kiwisoft.sqlPlugin.config;

import java.util.*;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyChangeListener;

import org.jdom.Element;

import com.intellij.openapi.application.Application;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.components.ApplicationComponent;
import com.intellij.openapi.util.InvalidDataException;
import com.intellij.openapi.util.JDOMExternalizable;
import com.intellij.openapi.util.WriteExternalException;

import com.kiwisoft.db.driver.DatabaseDriverManager;
import com.kiwisoft.db.DatabaseManager;
import com.kiwisoft.sqlPlugin.templates.TemplateManager;
import com.kiwisoft.sqlPlugin.JdbcLibrary;
import com.kiwisoft.utils.Utils;

/**
 * @author Stefan Stiller
 * @version $Revision: 1.3 $, $Date: 2006/02/18 15:31:53 $
 */
public class SQLPluginAppConfig implements ApplicationComponent, JDOMExternalizable, SQLPluginConstants
{

	private static SQLPluginAppConfig instance;

	public static SQLPluginAppConfig getInstance()
	{
		Application application=ApplicationManager.getApplication();
		if (application!=null)
		{
			Object component=application.getComponent(SQLPluginAppConfig.class);
			return (SQLPluginAppConfig)component;
		}
		else
		{
			if (instance==null)
			{
				instance=new SQLPluginAppConfig();
				instance.initComponent();
			}
			return instance;
		}
	}

	private ExportConfiguration export;

	public ExportConfiguration getExportConfiguration()
	{
		if (export==null) export=new ExportConfiguration();
		return export;
	}

	private TableConfigurationAdapter tableAdapter;

	public TableConfigurationAdapter getTableAdapter()
	{
		if (tableAdapter==null) tableAdapter=new TableConfigurationAdapter();
		return tableAdapter;
	}

	private Set jdbcLibraries=Collections.EMPTY_SET;

	public void setJdbcLibraries(Set urls)
	{
		if (!Utils.equals(jdbcLibraries, urls))
		{
			if (urls==null) jdbcLibraries=Collections.EMPTY_SET;
			else jdbcLibraries=new HashSet(urls);
			DatabaseDriverManager.reloadDrivers(jdbcLibraries);
		}
	}

	public Set getJdbcLibraries()
	{
		return Collections.unmodifiableSet(jdbcLibraries);
	}

	private boolean keepConnectionOpen;

	public boolean isKeepConnectionOpen()
	{
		return keepConnectionOpen;
	}

	public void setKeepConnectionOpen(boolean keepConnectionOpen)
	{
		this.keepConnectionOpen=keepConnectionOpen;
	}

	private boolean saveQueries;
	private String queryPath;
	private boolean rowLimitEnabled=true;
	private int rowLimit=1000;
	private boolean showPrimaryKeyColumns=true;
	private boolean savePasswords;
	private boolean encodePasswords=true;
	private boolean resizeColumnsToHeader=true;
	private boolean resizeColumnsToContent=true;
	private boolean saveResultTableConfiguration;
	private boolean includeProjectClasses;
	private String nullString;
	private Map confirmations=new HashMap();
	private Map defaultFormats=new HashMap();
	private boolean loadLOBs;
	private boolean stopOnError;

	public boolean isIncludeProjectClasses()
	{
		return includeProjectClasses;
	}

	public void setIncludeProjectClasses(boolean newValue)
	{
		boolean oldValue=includeProjectClasses;
		includeProjectClasses=newValue;
		propertyChangeSupport.firePropertyChange("includePropertyClasses", oldValue, newValue);
	}

	public String getNullString()
	{
		return nullString;
	}

	public void setNullString(String newValue)
	{
		String oldValue=nullString;
		nullString=newValue;
		propertyChangeSupport.firePropertyChange("nullString", oldValue, newValue);
	}

	public Boolean getConfirmation(String key)
	{
		return (Boolean)confirmations.get(key);
	}

	public void setConfirmation(String key, Boolean newValue)
	{
		Boolean oldValue=(Boolean)confirmations.put(key, newValue);
		propertyChangeSupport.firePropertyChange(key, oldValue, newValue);
	}

	public int getRowLimit()
	{
		return rowLimit;
	}

	public void setRowLimit(int newValue)
	{
		int oldValue=rowLimit;
		if (newValue<0) this.rowLimit=1000;
		this.rowLimit=newValue;
		propertyChangeSupport.firePropertyChange("rowLimit", oldValue, rowLimit);
	}

	public void setRowLimitEnabled(boolean newValue)
	{
		boolean oldValue=rowLimitEnabled;
		this.rowLimitEnabled=newValue;
		propertyChangeSupport.firePropertyChange("rowLimitEnabled", oldValue, rowLimitEnabled);
	}

	public boolean isRowLimitEnabled()
	{
		return rowLimitEnabled;
	}

	public boolean isSaveResultTableConfiguration()
	{
		return saveResultTableConfiguration;
	}

	public void setSaveResultTableConfiguration(boolean saveResultTableConfiguration)
	{
		this.saveResultTableConfiguration=saveResultTableConfiguration;
	}

	public void setDefaultFormat(Class aClass, String format)
	{
		String oldFormat=getDefaultFormat(aClass);
		defaultFormats.put(aClass.getName(), format);
		propertyChangeSupport.firePropertyChange("defaultFormat", oldFormat, format);
	}

	public String getDefaultFormat(Class aClass)
	{
		return (String)defaultFormats.get(aClass.getName());
	}

	public boolean isSaveQueries()
	{
		return saveQueries;
	}

	public void setSaveQueries(boolean saveQueries)
	{
		this.saveQueries=saveQueries;
	}

	public boolean isShowPrimaryKeyColumns()
	{
		return showPrimaryKeyColumns;
	}

	public void setShowPrimaryKeyColumns(boolean showPrimaryKeyColumns)
	{
		this.showPrimaryKeyColumns=showPrimaryKeyColumns;
	}

	public boolean isSavePasswords()
	{
		return savePasswords;
	}

	public void setSavePasswords(boolean savePasswords)
	{
		this.savePasswords=savePasswords;
	}

	public boolean isEncodePasswords()
	{
		return encodePasswords;
	}

	public void setEncodePasswords(boolean encodePasswords)
	{
		this.encodePasswords=encodePasswords;
	}

	public void setDriversPath(String driversPath)
	{
	}

	public String getQueryPath()
	{
		return queryPath;
	}

	public void setQueryPath(String queryPath)
	{
		this.queryPath=queryPath;
	}

	public boolean isResizeColumnsToHeader()
	{
		return resizeColumnsToHeader;
	}

	public void setResizeColumnsToHeader(boolean resizeColumnsToHeader)
	{
		this.resizeColumnsToHeader=resizeColumnsToHeader;
	}

	public boolean isResizeColumnsToContent()
	{
		return resizeColumnsToContent;
	}

	public void setResizeColumnsToContent(boolean resizeColumnsToContent)
	{
		this.resizeColumnsToContent=resizeColumnsToContent;
	}

	public boolean isLoadLargeObjects()
	{
		return loadLOBs;
	}

	public void setLoadLargeObjects(boolean newValue)
	{
		loadLOBs=newValue;
	}

	public boolean isStopOnError()
	{
		return stopOnError;
	}

	public void setStopOnError(boolean stopOnError)
	{
		this.stopOnError=stopOnError;
	}


	// Listener Support

	private PropertyChangeSupport propertyChangeSupport=new PropertyChangeSupport(this);

	public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener)
	{
		propertyChangeSupport.addPropertyChangeListener(propertyName, listener);
	}

	public void addPropertyChangeListener(PropertyChangeListener listener)
	{
		propertyChangeSupport.addPropertyChangeListener(listener);
	}

	public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener)
	{
		propertyChangeSupport.removePropertyChangeListener(propertyName, listener);
	}

	public void removePropertyChangeListener(PropertyChangeListener listener)
	{
		propertyChangeSupport.removePropertyChangeListener(listener);
	}


	// ApplicationComponent interface

	public String getComponentName()
	{
		return "SQLPlugin.Configuration";
	}

	public void initComponent()
	{
	}

	public void disposeComponent()
	{
		DatabaseManager.closeApplicationInstance();
	}

	// JDOMExternalizable interface

	public void writeExternal(Element element) throws WriteExternalException
	{
		try
		{
			// General options
			Element generalElement=new Element(GENERAL);
			Utils.setValue(generalElement, CONFIRM_COMMIT, getConfirmation(CONFIRM_COMMIT));
			Utils.setValue(generalElement, CONFIRM_ROLLBACK, getConfirmation(CONFIRM_ROLLBACK));
			Utils.setValue(generalElement, CONFIRM_DISCONNECT, getConfirmation(CONFIRM_DISCONNECT));
			generalElement.setAttribute(SAVE_PASSWORDS, Boolean.toString(isSavePasswords()));
			generalElement.setAttribute(ENCODE_PASSWORDS, Boolean.toString(isEncodePasswords()));
			generalElement.setAttribute(KEEP_CONNECTIONS_OPEN, Boolean.toString(keepConnectionOpen));
			element.addContent(generalElement);

			// Export Options
			Element exportElement=new Element(EXPORT);
			element.addContent(exportElement);
			getExportConfiguration().writeExternal(element);

			// JDBC Drivers
			Element driversElement=new Element(DRIVERS);
			for (Iterator it=jdbcLibraries.iterator(); it.hasNext();)
			{
				String url=((JdbcLibrary)it.next()).getVfsUrl();
				Element driverElement=new Element(DRIVER);
				driverElement.setAttribute(URL, url);
				driversElement.addContent(driverElement);
			}
			element.addContent(driversElement);

			// Databases
			Element databasesElement=new Element(DATABASES);
			DatabaseManager.getApplicationInstance().writeExternal(databasesElement);
			element.addContent(databasesElement);

			// Queries
			Element child=new Element(QUERIES);
			Utils.setValue(child, PATH, getQueryPath());
			Utils.setValue(child, PARTIAL_EXECUTE, getConfirmation(PARTIAL_EXECUTE));
			child.setAttribute(PROJECT_CLASS_LOADER, Boolean.toString(isIncludeProjectClasses()));
			child.setAttribute(SAVE, Boolean.toString(isSaveQueries()));
			child.setAttribute(STOP_ON_ERROR, Boolean.toString(isStopOnError()));
			element.addContent(child);

			// Results
			Element resultsElement=new Element(RESULTS);
			resultsElement.setAttribute(PRIMARY_KEY_COLUMNS, Boolean.toString(isShowPrimaryKeyColumns()));
			resultsElement.setAttribute(LOAD_LOBS, Boolean.toString(isShowPrimaryKeyColumns()));
			resultsElement.setAttribute(SAVE_TABLE_CONFIG, Boolean.toString(isSaveResultTableConfiguration()));
			for (Iterator it=defaultFormats.keySet().iterator(); it.hasNext();)
			{
				String className=(String)it.next();
				Element formatElement=new Element(DEFAULT_FORMAT);
				Utils.setValue(formatElement, CLASS, className);
				Utils.setValue(formatElement, FORMAT, defaultFormats.get(className));
				resultsElement.addContent(formatElement);
			}
			Utils.setValue(resultsElement, NULL_STRING, getNullString());
			element.addContent(resultsElement);

			Element rowLimitElement=new Element(ROW_LIMIT);
			rowLimitElement.setAttribute(ENABLED, Boolean.toString(isRowLimitEnabled()));
			rowLimitElement.setAttribute(VALUE, Integer.toString(getRowLimit()));
			resultsElement.addContent(rowLimitElement);

			Element resizeColumnsElement=new Element(RESIZE_COLUMNS);
			resizeColumnsElement.setAttribute(TO_HEADER, Boolean.toString(isResizeColumnsToHeader()));
			resizeColumnsElement.setAttribute(TO_CONTENT, Boolean.toString(isResizeColumnsToContent()));
			resultsElement.addContent(resizeColumnsElement);

			// Tables
			Element tablesElement=new Element(TABLES);
			element.addContent(tablesElement);
			getTableAdapter().writeExternal(tablesElement);

			if (TemplateManager.hasInstance()) TemplateManager.getInstance().saveTemplates();
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}

	public void readExternal(Element element) throws InvalidDataException
	{
		try
		{
			// General Options
			Element generalElement=element.getChild(GENERAL);
			if (generalElement!=null)
			{
				setConfirmation(CONFIRM_COMMIT, Utils.getBoolean(generalElement, CONFIRM_COMMIT, null));
				setConfirmation(CONFIRM_ROLLBACK, Utils.getBoolean(generalElement, CONFIRM_ROLLBACK, null));
				setConfirmation(CONFIRM_DISCONNECT, Utils.getBoolean(generalElement, CONFIRM_DISCONNECT, null));
				setSavePasswords(Utils.getBoolean(generalElement, SAVE_PASSWORDS, false));
				setEncodePasswords(Utils.getBoolean(generalElement, ENCODE_PASSWORDS, true));
				setKeepConnectionOpen(Utils.getBoolean(generalElement, KEEP_CONNECTIONS_OPEN, false));
			}

			// Export Options
			Element exportElement=element.getChild(EXPORT);
			if (exportElement!=null) getExportConfiguration().readExternal(exportElement);

			// JDBC Drivers
			Element driversElement=element.getChild(DRIVERS);
			Set libraries=new HashSet();
			if (driversElement!=null)
			{
				List children=driversElement.getChildren(DRIVER);
				for (Iterator it=children.iterator(); it.hasNext();)
				{
					Element driverElement=(Element)it.next();
					String url=Utils.getString(driverElement, URL, null);
					if (url!=null)
					{
						libraries.add(new JdbcLibrary(url));
					}
				}
			}
			setJdbcLibraries(libraries);

			// Databases
			Element databasesElement=element.getChild(DATABASES);
			if (databasesElement!=null)
			{
				DatabaseManager.getApplicationInstance().readExternal(databasesElement, isEncodePasswords());
			}

			// Queries
			Element queriesElement=element.getChild(QUERIES);
			if (queriesElement!=null)
			{
				setConfirmation(PARTIAL_EXECUTE, Utils.getBoolean(queriesElement, PARTIAL_EXECUTE, null));
				setQueryPath(queriesElement.getAttributeValue(PATH));
				setSaveQueries(Utils.getBoolean(queriesElement, SAVE, false));
				setIncludeProjectClasses(Utils.getBoolean(queriesElement, PROJECT_CLASS_LOADER, false));
				setStopOnError(Utils.getBoolean(queriesElement, STOP_ON_ERROR, true));
			}

			// Results
			Element resultsElement=element.getChild(RESULTS);
			if (resultsElement!=null)
			{
				setShowPrimaryKeyColumns(Utils.getBoolean(resultsElement, PRIMARY_KEY_COLUMNS, true));
				setLoadLargeObjects(Utils.getBoolean(resultsElement, LOAD_LOBS, true));
				setSaveResultTableConfiguration(Utils.getBoolean(resultsElement, SAVE_TABLE_CONFIG, false));
				List children=resultsElement.getChildren(DEFAULT_FORMAT);
				if (children!=null)
				{
					for (Iterator it=children.iterator(); it.hasNext();)
					{
						Element formatElement=(Element)it.next();
						String className=formatElement.getAttributeValue(CLASS);
						String format=formatElement.getAttributeValue(FORMAT);
						defaultFormats.put(className, format);
					}
				}
				setNullString(resultsElement.getAttributeValue(NULL_STRING));

				Element rowLimitElement=resultsElement.getChild(ROW_LIMIT);
				setRowLimitEnabled(Utils.getBoolean(rowLimitElement, ENABLED, true));
				setRowLimit(Utils.getInteger(rowLimitElement, VALUE, new Integer(1000)).intValue());

				Element resizeElement=element.getChild(RESIZE_COLUMNS);
				if (resizeElement!=null)
				{
					setResizeColumnsToHeader(Utils.getBoolean(resizeElement, TO_HEADER, true));
					setResizeColumnsToContent(Utils.getBoolean(resizeElement, TO_CONTENT, true));
				}
			}

			// Tables
			Element tablesElement=element.getChild(TABLES);
			if (tablesElement!=null) getTableAdapter().readExternal(tablesElement);
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}

	}
}
